#!/usr/bin/env python
import re, os, sys,time
import logging
import numpy

import logConfig
import optParser
import fileParser
import shiftSize
import windowSize
import sigTests
import misc

# initialize the logger
root_logger = logging.getLogger("")
debug = root_logger.debug
info = root_logger.info


def main(argv):
        ## performing the option parser
    opt = optParser.opt_parser(argv)
    parameter, readData = optParser.process_opt(opt)
    ## read in the data
    fileParser.parse(readData, parameter.file_format)

    ## remove the redundant reads
    if (parameter.remove_redundant):
        readData.remove_redundant_reads()

    ## shiftSize estimation and shifting reads
    shiftSize.estimate_shift_size(readData,parameter)
    shiftSize.shift_reads(readData)

    file_in = open(argv[1], 'r')
    file_passed = open(argv[1]+'_passed', 'w')
    file_filtered = open(argv[1]+'_filtered', 'w')
    post_processing(readData,file_in,file_passed,file_filtered)

def post_processing(readData, peak, passed, filtered, remove_artefacts=True, narrow_peak=False):
    info(" Begin post-processing.")
    chip_list =  readData.chip_filename_list
    input_list = readData.input_filename_list
    strands_dict = readData.data_dict_by_strands

    for line in peak: 
        words = line.strip().split('\t')
        chr = words[0]
        start = int(words[1])
        end = int(words[2])
        rest = words[3:]
        (start, end, chip_input_ratio, tag_monopoly, overlap_orig,
        overlap_roll) = post_processing_per_peak(
            strands_dict, chip_list, input_list, chr, start, end,
            readData.shift_size, readData.read_length,
            narrow_peak, remove_artefacts)
        if remove_artefacts and (chip_input_ratio > 0.5 or
                # tag_monopoly > 0.5 or 
                (overlap_orig > 0.2 and
                overlap_roll/overlap_orig < 0.5)):
            filtered.write(chr+"\t" + str(start) + '\t' + str(end) + '\t' +
                        str(end-start)+'\t')
            for item in rest: 
                filtered.write(item +'\t')
            filtered.write(str(chip_input_ratio) + '\t' +
                        str(tag_monopoly) + '\t' + str(overlap_orig) +
                        '\t' + str(overlap_roll) + '\n')
            continue
        passed.write(chr+"\t" + str(start) + '\t' + str(end) + '\t' +
                     str(end-start)+'\t')
        for item in rest:
            passed.write(item +'\t')
        passed.write(str(chip_input_ratio) + '\t' +
                         str(tag_monopoly) + '\t' + str(overlap_orig) +
                         '\t' + str(overlap_roll) + '\n')
    return 


def post_processing_per_peak(strands_dict, chip_list, input_list, chr,
                             start, end, shiftSize, readLength, narrow_peak,
                             remove_artefacts):
    ''' Remove artefacts and refine peak width.'''
    chip_forward = numpy.zeros(end-start)
    chip_reverse = numpy.zeros(end-start)
    input_forward = numpy.zeros(end-start)
    input_reverse = numpy.zeros(end-start)
    for chip in chip_list:
        forward = strands_dict[chr][chip]['f']
        reverse = strands_dict[chr][chip]['r']
        forward_read = forward[numpy.where( (forward >= start) &
                (forward < end) )]
        reverse_read = reverse[numpy.where( (reverse >= start+readLength) &
                (reverse < end+readLength) )]
        for read in forward_read:
            try: chip_forward[(read-start)] +=1
            except IndexError:
                debug("index error ignored. end-start: %d. read-start: %d.", end-start, read-start)
        for read in reverse_read:
            try: chip_reverse[(read-start-readLength)] +=1
            except IndexError:
                debug("index error ignored. end-start: %d. read-start-readLength: %d.", end-start, read-start-readLength)
    # using input reads to remove artefacts
    if remove_artefacts is True:
        for input in input_list:
            forward = strands_dict[chr][input]['f']
            reverse = strands_dict[chr][input]['r']
            forward_read = forward[numpy.where( (forward >=
                    start) & (forward < end) )]
            reverse_read = reverse[numpy.where( (reverse >=
                    start+readLength) & (reverse < end+readLength) )]
            for read in forward_read:
                try: input_forward[(read-start)] +=1
                except IndexError:
                    debug("index error ignored. end-start: %d. read-start: %d.", end-start, read-start)
            for read in reverse_read:
                try: input_reverse[(read-start-readLength)] +=1
                except IndexError:
                    debug("index error ignored. end-start: %d. read-start-readLength: %d.", end-start, read-start-readLength)

        chip_both = chip_forward + chip_reverse
        input_both = input_forward + input_reverse
        if sum(chip_both) == 0:
            pass
        else:
            chip_both = chip_both/sum(chip_both)
        if sum(input_both) ==0:
            pass
        else:
            input_both = input_both/sum(input_both)
        overlap_chip_input = numpy.sum(numpy.minimum(chip_both, input_both))
        chip_both.sort()
        chip_3_maximum = numpy.sum(chip_both[-3:])
        if sum(chip_reverse) != 0:
            chip_reverse = chip_reverse/sum(chip_reverse)
        chip_forward_roll = numpy.roll(chip_forward,readLength)
        if sum(chip_forward) != 0:
            chip_forward = chip_forward/sum(chip_forward)
            chip_forward_roll = chip_forward_roll/sum(chip_forward)
        overlap_orig = numpy.sum(numpy.minimum(chip_forward, chip_reverse))
        overlap_orig = numpy.max([overlap_orig, 1e-5])
        overlap_roll = numpy.sum(numpy.minimum(chip_forward_roll, chip_reverse))
    else:
        overlap_chip_input = 0
        chip_3_maximum = 0
        overlap_orig = 1e-5
        overlap_roll = 0

    if narrow_peak is True:
        sum_forward = 0
        sum_reverse = 0
        if sum(chip_forward) > 0:
            chip_forward = chip_forward/sum(chip_forward)
            for i in xrange(end-start):
                sum_forward += chip_forward[i]
                if sum_forward > 0.2:
                    new_start = start + i
                    break
        if sum(chip_reverse) > 0:
            chip_reverse = chip_reverse/sum(chip_reverse)
            for i in xrange(end-start-1, -1, -1):
                sum_reverse += chip_reverse[i]
                if sum_reverse > 0.2:
                    new_end = start + i
                    break
        if sum(chip_forward) == 0:
            new_start = new_end - 2*shiftSize
        if sum(chip_reverse) == 0:
            new_end = new_start + 2*shiftSize
        start = new_start
        end = new_end

    return (start, end, overlap_chip_input, chip_3_maximum,
            overlap_orig, overlap_roll)



if __name__ == '__main__':
    try: main(sys.argv)
    except KeyboardInterrupt:
        print "user interrupted me"





#!/usr/bin/env python26

import sys
import re
import logging

import numpy
import pysam


root_logger = logging.getLogger("")
info = root_logger.info


def bam_parse(filename_list):
    '''Parsing bam format files.'''
    data_dict = {}
    infile = pysam.Samfile(filename_list[0], 'rb')
    line = infile.fetch().next()
    length = line.alen
    infile.close()
    info("the length of the reads is %s", length)

    for filename in filename_list: 
        infile = pysam.Samfile(filename, 'rb')
        info("retrieving reads from file : %s", filename)
        for line in infile.fetch():
            if line.is_unmapped is False: 
                chr = infile.getrname(line.tid)
                pos = line.pos
                if line.is_reverse is False: 
                    try:
                        data_dict[chr][filename]['f'].append(pos)
                    except KeyError:
                        if chr not in data_dict:
                            data_dict[chr] = {}
                        data_dict[chr][filename] = {}
                        data_dict[chr][filename]['f'] =[pos]
                        data_dict[chr][filename]['r'] =[]
                else:
                    try:
                        data_dict[chr][filename]['r'].append(pos + length +1)
                    except KeyError:
                        if chr not in data_dict:
                            data_dict[chr] = {}
                        data_dict[chr][filename] = {}
                        data_dict[chr][filename]['f'] =[]
                        data_dict[chr][filename]['r'] =[pos + length + 1]
        infile.close()
    return data_dict, length



def sam_parse(filename_list):
    '''Parsing sam format files.'''
    # Store the position and strand data for each file,
    # seperated by chromosomes.
    data_dict = {}
    infile = open(filename_list[0], 'r')

    for line in infile:
        words = line.strip().split()
        if not words[0].startswith("@"):
            length = len(words[9])
            break
    infile.close()
    info("the length of the reads is %s", length)

    for filename in filename_list:
        infile = open(filename, 'r')
        info("retrieving reads from file : %s", filename)
        for line in infile: 
            if not line.startswith("@"):
                break
        for line in infile:
            words = line.strip().split()
            words[1] = int(words[1])

            if words[1] is not 0x0004:
                chr = words[2]
                pos = int(words[3])-1

                if words[1] is not 16:
                    try: 
                        data_dict[chr][filename]['f'].append(pos)
                    except KeyError:
                        if chr not in data_dict:
                            data_dict[chr] = {}
                        data_dict[chr][filename] = {}
                        data_dict[chr][filename]['f'] =[pos]
                        data_dict[chr][filename]['r'] =[]
                else:
                    try:
                        data_dict[chr][filename]['r'].append(pos + length +1)
                    except KeyError:
                        if chr not in data_dict:
                            data_dict[chr] = {}
                        data_dict[chr][filename] = {}
                        data_dict[chr][filename]['f'] =[]
                        data_dict[chr][filename]['r'] =[pos + length + 1]

                    except KeyError: pass
        infile.close()
    return data_dict, length

def bowtie_parse(filename_list, chr_list):
    #parsing bowtie format files
    data_dict = {} 
    infile = open(filename_list[0], 'r')
    line = infile.readline()
    words = line.strip().split()
    length = len(words[4])
    info("the length of the reads is %s", length)
    infile.close()

    for filename in filename_list:
        infile = open(filename, 'r')
        info("retrieving reads from file : %s", filename)

        for line in infile:
            line = line.strip().split()
            chr = line[2]
            strand = line[1]
            pos = int(line[3])

            if strand == '+':
                try: data_dict[chr][filename]['f'].append(pos)
                except KeyError:
                    if chr not in data_dict:
                        data_dict[chr] = {}
                    data_dict[chr][filename] = {}
                    data_dict[chr][filename]['f'] =[pos]
                    data_dict[chr][filename]['r'] =[]
            elif strand == '-':
                try: data_dict[chr][filename]['r'].append(pos + length)
                except KeyError:
                    if chr not in data_dict:
                        data_dict[chr] = {}
                    data_dict[chr][filename] = {}
                    data_dict[chr][filename]['f'] =[]
                    data_dict[chr][filename]['r'] =[pos + length + 1]
            else:
                print("strand error")
        infile.close()
    return data_dict, length

def bed_parse(filename_list):
    '''Parsing BED format files'''
    data_dict={} 
    for filename in filename_list:
	print filename
        infile = open(filename, 'r')
        info("retrieving reads from file: %s", filename)
        for line in infile:
            line = line.strip().split()
            chr=line[0]
            strand = line[5]
            if strand == '+':
                pos = int(line[1])  # genomic position
                try: data_dict[chr][filename]['f'].append(pos)
                except KeyError:
                    if chr not in data_dict:
                        data_dict[chr] = {}
                    data_dict[chr][filename] = {}
                    data_dict[chr][filename]['f'] =[pos]
                    data_dict[chr][filename]['r'] =[]
            elif strand == '-':
                pos = int(line[2])-1
                try: data_dict[chr][filename]['r'].append(pos)
                except KeyError:
                    if chr not in data_dict:
                        data_dict[chr] = {}
                    data_dict[chr][filename] = {}
                    data_dict[chr][filename]['f'] =[]
                    data_dict[chr][filename]['r'] =[pos]
            else:
                print("strand error")
        infile.close()
    return data_dict, int(line[2])-int(line[1])  # read_length

def process_illumina_match(align, mismatch, length, format):
    words = align.split(',')
    if format == "multi":
        for match in words:
            if match.startswith('c'):
                str, match =match.split(':')
                chr = re.match(r'(chr(\d\d?|\w))', str)
                chr = chr.group(0)
            if match.endswith(mismatch):
                strand = match[-2]
                pos = int(match[0:-2])
                if strand =='R':
                    pos = pos+length
                return chr, strand, pos
    elif format == "extended":
        for match in words:
            if match.startswith('c'):
                chr_part, match = match.split(':')
                chr = re.match(r'(chr(\d\d?|\w))', chr_part)
                chr = chr.group(0)
                m1 = re.search(r'([F|R])', match)
                strand = m1.group(1)
                pos = int(re.sub(r'[F|R].*', '', match))

                if strand == 'R':
                    pos = pos + length

                match = re.sub(r'(\^\w+\$)', '', match)
                m2 = re.findall(r'[A-Z]', match)

                if len(m2) - 1 == int(mismatch):
                    return chr, strand, pos


# Parsing eland format files
def eland_parse(filename_list, chr_list, format="default"):
    #store the position and strand data for each file, seperated by chromosomes.
    data_dict={}
    infile = open(filename_list[0], 'r')
    line = infile.readline()
    words=line.split()
    length=len(words[1])
    info ("the length of the reads is %s", length)
    infile.close()

    for filename in filename_list:
        infile = open(filename, 'r')
        info("retrieving reads from file:%s", filename)
        for line in infile:
            words = line.strip().split()
            mismatch = -1
            
            if len(words) == 4 and ':' in words[2]:
                num_mismatch = words[2].split(':')
                
                if num_mismatch[0]== '1':
                    mismatch = '0'
                elif (num_mismatch[0]=='0') and (num_mismatch[1] == '1'):
                    mismatch = '1'
                elif (num_mismatch[0]=='0' and num_mismatch[1]=='0' and
                        num_mismatch[2] == '1'):
                    mismatch = '2'
                if mismatch !=-1:
                    try:
                        chr, strand, pos = process_illumina_match(
                                words[3], mismatch, length, format)
                        if strand == 'F':
                            try: data_dict[chr][filename]['f'].append(pos)
                            except KeyError:
                                if chr not in data_dict:
                                    data_dict[chr] = {}
                                data_dict[chr][filename] = {}
                                data_dict[chr][filename]['f'] =[pos]
                                data_dict[chr][filename]['r'] =[]
                        elif strand == 'R':
                            try: data_dict[chr][filename]['r'].append(pos)
                            except KeyError:
                                if chr not in data_dict:
                                    data_dict[chr] = {}
                                data_dict[chr][filename] = {}
                                data_dict[chr][filename]['f'] =[]
                                data_dict[chr][filename]['r'] =[pos + length + 1]
                    except TypeError:
                        pass
        infile.close()
    return data_dict, length


def parse(readData, file_format):
    '''retrieve the data form files and make it into appropriate data 
        format'''
    info("begin parsing files...")
    data_dict = {}
    if file_format == "bed":
        data_dict, read_length = bed_parse(readData.filename_list)
    elif file_format == "eland_multi":
        data_dict, read_length = eland_parse(
                readData.filename_list, readData.chr_list, format="multi")
    elif file_format == "eland_extended":
        data_dict, read_length = eland_parse(
                readData.filename_list, readData.chr_list, format="extended")
    elif file_format == "bowtie":
        data_dict, read_length = bowtie_parse(
                readData.filename_list, readData.chr_list)
    elif file_format == "sam":
        data_dict, read_length = sam_parse(readData.filename_list)
    elif file_format == "bam":
        data_dict, read_length = bam_parse(readData.filename_list)
    else: 
        pass

    #--- here can add other file format support ---#

    # fill in the data structures. 
    readData.data_dict_by_strands = data_dict
    readData.read_length = read_length
    readData.refine_chr_list()
    readData.cal_read_total()
    for file in readData.filename_list: 
        info("Total #reads for %s is %s", file, 
                readData.read_total_per_file[file])
    readData.cal_genome_size()
    info( "leaving file_parser.py")

